import os
import cv2
import glob
import tqdm
import argparse
import time


def play_image(image_file_list, fps):
    period = 1.0 / fps
    for image_file in tqdm.tqdm(image_file_list):
        start_time = time.time()
        image = cv2.imread(image_file)
        cv2.namedWindow("image", 0)
        cv2.imshow("image", image)
        current_time = time.time()
        delta_time = current_time - start_time
        sleep_time = max(1, int((period - delta_time) * 1000))
        cv2.waitKey(sleep_time)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", type=str, default=".", help="path of image files")
    parser.add_argument(
        "-p",
        "--postfix",
        type=str,
        default="png",
        help="postfix of image filename",
    )
    parser.add_argument("-f", "--fps", type=float, default=29.97, help="desired fps for video")
    opts = parser.parse_args()
    print(opts)

    fps = opts.fps
    image_file_list = sorted(glob.glob(os.path.join(opts.input, f"*.{opts.postfix}")))
    play_image(image_file_list, fps)


if __name__ == "__main__":
    main()



# -*- encoding: utf-8 -*-
"""
@File    :   iterator.py
@Time    :   2023/03/14 10:00:03
@Author  :   lihao57
@Version :   1.0
@Contact :   lihao57@baidu.com
"""

import threading
import multiprocessing
from concurrent.futures import ThreadPoolExecutor, as_completed
import tqdm
import tqdm.asyncio
import functools
import asyncio
import aiomultiprocess
import numpy as np
import time
import inspect
from typing import Callable, Optional, Type


class Iterator:
    """
    Iterator

    Parameters:
        func (callable): function to be iterated
        total (int): number of iterations, if not specified, will infer from the list-type `args` and `kwargs` of `run` method
        disable_pbar (bool): whether to disable progress bar, default is `False`
    """

    def __init__(
        self,
        func: Callable,
        total: Optional[int] = None,
        disable_pbar: bool = False,
        **kwargs,
    ):
        self.func = func
        self.total = total
        self.disable_pbar = disable_pbar
        self.dynamic_args = None
        self.static_args = None
        self.dynamic_kwargs = None
        self.static_kwargs = None
        self.partial_func = None

    def parse(self, *args, **kwargs):
        """parse"""
        # infer the number of expected iterations from args
        if self.total is None:
            for arg in args:
                if isinstance(arg, list):
                    self.total = len(arg)
                    break

        # infer the number of expected iterations from kwargs
        if self.total is None:
            for _, value in kwargs.items():
                if isinstance(value, list):
                    self.total = len(value)
                    break

        assert self.total is not None, "Can not infer the number of expected iterations"

        # parse args
        self.dynamic_args = [[] for _ in range(self.total)]
        self.static_args = []
        dynamic_arg_flags = []
        for arg in args:
            if isinstance(arg, list):
                assert len(arg) == self.total, f"{len(arg)} != {self.total}"
                for i, val in enumerate(arg):
                    self.dynamic_args[i].append(val)
                dynamic_arg_flags.append(True)
            else:
                self.static_args.append(arg)
                dynamic_arg_flags.append(False)

        # Assert static arguments precede dynamic arguments
        if len(dynamic_arg_flags) > 1:
            dynamic_arg_flags = np.array(dynamic_arg_flags)
            if (dynamic_arg_flags[:-1] > dynamic_arg_flags[1:]).any():
                msg = ["dynamic" if flag else "static" for flag in dynamic_arg_flags]
                raise RuntimeError(f"Dynamic arguments precede static arguments: {msg}")

        # parse kwargs
        self.dynamic_kwargs = [{} for _ in range(self.total)]
        self.static_kwargs = {}
        for key, value in kwargs.items():
            if isinstance(value, list):
                assert len(value) == self.total, f"{len(value)} != {self.total}"
                for i, val in enumerate(value):
                    self.dynamic_kwargs[i][key] = val
            else:
                self.static_kwargs[key] = value

        # fix the static args
        self.partial_func = functools.partial(self.func, *self.static_args, **self.static_kwargs)

    def run(self, *args, **kwargs):
        """run"""
        raise NotImplementedError


class AutoIterator(Iterator):
    """
    Auto Iterator

    Create an iterator automatically based on the input arguments.

    Parameters:
        iterator_cls (type | None): iterator class, if not specified, will not use iterator
        func (callable): function to be iterated
        total (int): number of iterations, if not specified, will infer from the list-type `args` and `kwargs` of `run` method
        **kwargs: inherited from `Iterator`
    """

    def __init__(
        self,
        iterator_cls: Optional[Type[Iterator]],
        func: Callable,
        total: Optional[int] = None,
        **kwargs,
    ):
        if iterator_cls:
            self._func = iterator_cls(func=func, total=total, **kwargs).run
        else:
            self._func = func

    def run(self, *args, **kwargs):
        """run"""
        return self._func(*args, **kwargs)


class SingleProcess(Iterator):
    """
    SingleProcess

    Parameters:
        inherited from `Iterator`

    Example:
        ```python
        def add(a, b):
            return a + b

        a = [1, 2]
        b = [3, 4]
        res = SingleProcess(add).run(a, b)
        print(res)
        # [4, 6]
        ```
    """

    def run(self, *args, **kwargs):
        """
        run
            - Please ensure that static arguments precede dynamic arguments
            - If the `total` is not specified, it will be inferred from the first list-type parameter
        """
        ret_list = []
        if "_counter" in kwargs:
            _counter = kwargs.pop("_counter")

            # parse
            self.parse(*args, **kwargs)

            # run
            for args, kwargs in zip(self.dynamic_args, self.dynamic_kwargs):
                ret_list.append(self.partial_func(*args, **kwargs))
                with _counter.get_lock():
                    _counter.value += 1
        else:
            # parse
            self.parse(*args, **kwargs)

            # run
            for args, kwargs in tqdm.tqdm(
                zip(self.dynamic_args, self.dynamic_kwargs),
                total=self.total,
                desc=self.func.__name__,
                disable=self.disable_pbar,
            ):
                ret_list.append(self.partial_func(*args, **kwargs))
        return ret_list


class MultiProcess(Iterator):
    """
    Multi-Process

    Parameters:
        func (callable): function to be iterated
        total (int): number of iterations, if not specified, will infer from the list-type `args` and `kwargs` of `run` method
        nprocs (int): number of processes, default is `multiprocessing.cpu_count()`
        use_pool (bool): use `multiprocessing.Pool` or `multiprocessing.Process`, default is True
        pbar_refresh_interval (float): interval of progress bar refreshing when `use_pool` is False, default is 1.0s
        **kwargs: inherited from `Iterator`

    Example:
        ```python
        def add(a, b):
            return a + b

        a = [1, 2]
        b = [3, 4]
        # use `multiprocessing.Pool`
        res = MultiProcess(add).run(a, b)

        # use `multiprocessing.Process`
        res = MultiProcess(add, use_pool=False).run(a, b)

        print(res)
        # [4, 6]
        ```
    """

    def __init__(
        self,
        func: Callable,
        total: Optional[int] = None,
        nprocs: int = multiprocessing.cpu_count(),
        use_pool: bool = True,
        pbar_refresh_interval: float = 1.0,
        **kwargs,
    ):
        super().__init__(func=func, total=total, **kwargs)

        self.nprocs = nprocs if nprocs > 0 else multiprocessing.cpu_count()
        self.use_pool = use_pool
        self.pbar_refresh_interval = pbar_refresh_interval

    def _func_pool(self, args):
        """function when using `multiprocessing.Pool`"""
        idx = args[2]
        return idx, self.partial_func(*args[0], **args[1])

    def _func(self, input_queue, output_queue, counter):
        """function when using `multiprocessing.Process`"""
        while True:
            args = input_queue.get()
            if args is None:
                break

            idx = args[2]
            try:
                result = self.partial_func(*args[0], **args[1])
                output_queue.put((idx, result, None))
            except Exception as e:
                output_queue.put((idx, None, e))
            with counter.get_lock():
                counter.value += 1

    def run(self, *args, **kwargs):
        """
        run
            - Please ensure that static arguments precede dynamic arguments
            - If the `total` is not specified, it will be inferred from the first list-type parameter
        """
        # parse
        self.parse(*args, **kwargs)

        if self.use_pool:  # using `multiprocessing.Pool`
            # run
            ret_list = [None] * self.total
            with multiprocessing.Pool(self.nprocs) as p:
                for idx, result in tqdm.tqdm(
                    p.imap_unordered(self._func_pool, zip(self.dynamic_args, self.dynamic_kwargs, range(self.total))),
                    total=self.total,
                    desc=self.func.__name__,
                    disable=self.disable_pbar,
                ):
                    ret_list[idx] = result
            return ret_list
        else:  # using `multiprocessing.Process`
            # initialize data queue
            input_queue = multiprocessing.Queue()
            output_queue = multiprocessing.Queue()
            for args in zip(self.dynamic_args, self.dynamic_kwargs, range(self.total)):
                input_queue.put(args)
            for _ in range(self.nprocs):
                input_queue.put(None)

            # initialize counter
            counter = multiprocessing.Value("i", 0)

            # run
            procs = []
            for _ in range(self.nprocs):
                p = multiprocessing.Process(target=self._func, args=(input_queue, output_queue, counter))
                procs.append(p)
                p.start()

            # display progress bar
            with tqdm.tqdm(total=self.total, desc=self.func.__name__, disable=self.disable_pbar) as pbar:

                def _update_pbar():
                    """update function"""
                    last_val = 0
                    while any(p.is_alive() for p in procs):
                        val = counter.value
                        delta = val - last_val
                        if delta > 0:
                            pbar.update(delta)
                            last_val = val
                        time.sleep(self.pbar_refresh_interval)
                    delta = counter.value - last_val
                    if delta > 0:
                        pbar.update(delta)

                t = threading.Thread(target=_update_pbar, daemon=True)
                t.start()

                # Get output
                ret_list = [None] * self.total
                for _ in range(self.total):
                    idx, result, e = output_queue.get()
                    if e is not None:
                        raise e  # TODO: terminate all processes
                    ret_list[idx] = result

                for p in procs:
                    p.join()
                t.join()

            return ret_list


class AutoMultiProcess(AutoIterator):
    """
    Auto Multi-Process

    Create a `MultiProcess` iterator when `nprocs > 1` and a `SingleProcess` iterator otherwise.

    Parameters:
        func (callable): function to be iterated
        total (int): number of iterations, if not specified, will infer from the list-type `args` and `kwargs` of `run` method
        nprocs (int): number of processes, default is `multiprocessing.cpu_count()`
        use_pool (bool): use `multiprocessing.Pool` or `multiprocessing.Process`, default is True
        pbar_refresh_interval (float): interval of progress bar refreshing when `use_pool` is False, default is 1.0s
        **kwargs: inherited from `AutoIterator`

    Example:
        ```python
        def add(a, b):
            return a + b

        a = [1, 2]
        b = [3, 4]
        # Use MultiProcess
        res = AutoMultiProcess(add, nprocs=2).run(a, b)
        print(res)
        # [4, 6]

        # Use SingleProcess
        res = AutoMultiProcess(add, nprocs=1).run(a, b)
        print(res)
        # [4, 6]
        ```
    """

    def __init__(
        self,
        func: Callable,
        total: Optional[int] = None,
        nprocs: int = multiprocessing.cpu_count(),
        use_pool: bool = True,
        pbar_refresh_interval: float = 1.0,
        **kwargs,
    ):
        iterator_cls = MultiProcess if nprocs > 1 else SingleProcess
        super().__init__(
            iterator_cls=iterator_cls,
            func=func,
            total=total,
            nprocs=nprocs,
            use_pool=use_pool,
            pbar_refresh_interval=pbar_refresh_interval,
            **kwargs,
        )


class BoundedMultiProcess(Iterator):
    """
    Bounded Multi-Process

    A utility to run a function in parallel using multiple processes,
    with support for per-process resource binding (e.g., GPU ID, port).
    Make sure the number of resources equals the number of processes,
    and the resources arguments must be passed as keyword arguments.

    Parameters:
        func (callable): function to be iterated
        total (int): number of iterations, if not specified, will infer from the list-type `args` and `kwargs` of `run` method
        nprocs (int): number of processes, default is `multiprocessing.cpu_count()`
        use_pool (bool): use `multiprocessing.Pool` or `multiprocessing.Process`, default is True
        pbar_refresh_interval (float): interval of progress bar refreshing when `use_pool` is False, default is 1.0s
        **kwargs: inherited from `Iterator`

    Example:
        ```python
        def add(a, b, port):
            print(f"{port}: {a} + {b}")
            return a + b

        a = [1, 2, 3, 4]
        b = [5, 6, 7, 8]
        # length of `port` must be equal to `nprocs`
        res = MultiProcess(add, nprocs=2).run(a, b, port=[8000, 8001])

        # use `multiprocessing.Process`
        res = MultiProcess(add, nprocs=2, use_pool=False).run(a, b, port=[8000, 8001])

        print(res)
        # [6, 8, 10, 12]
        ```
    """

    def __init__(
        self,
        func: Callable,
        total: Optional[int] = None,
        nprocs: int = multiprocessing.cpu_count(),
        use_pool: bool = True,
        pbar_refresh_interval: float = 1.0,
        **kwargs,
    ):
        super().__init__(func=func, total=total, **kwargs)

        self.nprocs = nprocs if nprocs > 0 else multiprocessing.cpu_count()
        self.use_pool = use_pool
        self.pbar_refresh_interval = pbar_refresh_interval

    def parse(self, *args, **kwargs):
        """parse"""
        # infer the number of expected iterations from args
        if self.total is None:
            for arg in args:
                if isinstance(arg, list):
                    self.total = len(arg)
                    break

        # infer the number of expected iterations from kwargs
        if self.total is None:
            for _, value in kwargs.items():
                if isinstance(value, list):
                    self.total = len(value)
                    break

        assert self.total is not None, "Can not infer the number of expected iterations"
        assert (
            self.total != self.nprocs
        ), "Ensure that `total != nprocs`; otherwise, it will be unclear whether the argument represents a resource. Please use `MultiProcess` instead"

        # parse args
        self.dynamic_args = [[] for _ in range(self.total)]
        self.static_args = []
        dynamic_arg_flags = []
        for arg in args:
            if isinstance(arg, list):
                assert len(arg) == self.total, f"{len(arg)} != {self.total}"
                for i, val in enumerate(arg):
                    self.dynamic_args[i].append(val)
                dynamic_arg_flags.append(True)
            else:
                self.static_args.append(arg)
                dynamic_arg_flags.append(False)

        # Assert static arguments precede dynamic arguments
        if len(dynamic_arg_flags) > 1:
            dynamic_arg_flags = np.array(dynamic_arg_flags)
            if (dynamic_arg_flags[:-1] > dynamic_arg_flags[1:]).any():
                msg = ["dynamic" if flag else "static" for flag in dynamic_arg_flags]
                raise RuntimeError(f"Dynamic arguments precede static arguments: {msg}")

        # parse kwargs
        self.process_kwargs = [{} for _ in range(self.nprocs)]
        self.dynamic_kwargs = [{} for _ in range(self.total)]
        self.static_kwargs = {}
        for key, value in kwargs.items():
            if isinstance(value, list):
                if len(value) == self.nprocs:
                    for i, val in enumerate(value):
                        self.process_kwargs[i][key] = val
                else:
                    assert len(value) == self.total, f"{len(value)} != {self.total}"
                    for i, val in enumerate(value):
                        self.dynamic_kwargs[i][key] = val
            else:
                self.static_kwargs[key] = value

        # fix the static args
        self.partial_func = functools.partial(self.func, *self.static_args, **self.static_kwargs)

    def initializer(self, counter):
        """initializer"""
        # Set process index
        global _proc_idx
        with counter.get_lock():
            _proc_idx = counter.value
            counter.value += 1

    def _func_pool(self, args):
        """function when using `multiprocessing.Pool`"""
        # Get process index
        global _proc_idx
        idx = args[2]
        return idx, self.partial_func(*args[0], **args[1], **self.process_kwargs[_proc_idx])

    def _func(self, input_queue, output_queue, counter, proc_idx):
        """function when using `multiprocessing.Process`"""
        while True:
            args = input_queue.get()
            if args is None:
                break

            idx = args[2]
            try:
                result = self.partial_func(*args[0], **args[1], **self.process_kwargs[proc_idx])
                output_queue.put((idx, result, None))
            except Exception as e:
                output_queue.put((idx, None, e))
            with counter.get_lock():
                counter.value += 1

    def run(self, *args, **kwargs):
        """
        run
            - Please ensure that static arguments precede dynamic arguments
            - If the `total` is not specified, it will be inferred from the first list-type parameter
        """
        # parse
        self.parse(*args, **kwargs)

        if self.use_pool:  # using `multiprocessing.Pool`
            # run
            counter = multiprocessing.Value("i", 0)
            ret_list = [None] * self.total
            with multiprocessing.Pool(self.nprocs, self.initializer, (counter,)) as p:
                for idx, result in tqdm.tqdm(
                    p.imap_unordered(self._func_pool, zip(self.dynamic_args, self.dynamic_kwargs, range(self.total))),
                    total=self.total,
                    desc=self.func.__name__,
                    disable=self.disable_pbar,
                ):
                    ret_list[idx] = result
            return ret_list
        else:  # using `multiprocessing.Process`
            # initialize data queue
            input_queue = multiprocessing.Queue()
            output_queue = multiprocessing.Queue()
            for args in zip(self.dynamic_args, self.dynamic_kwargs, range(self.total)):
                input_queue.put(args)
            for _ in range(self.nprocs):
                input_queue.put(None)

            # initialize counter
            counter = multiprocessing.Value("i", 0)

            # run
            procs = []
            for proc_idx in range(self.nprocs):
                p = multiprocessing.Process(target=self._func, args=(input_queue, output_queue, counter, proc_idx))
                procs.append(p)
                p.start()

            # display progress bar
            with tqdm.tqdm(total=self.total, desc=self.func.__name__, disable=self.disable_pbar) as pbar:

                def _update_pbar():
                    """update function"""
                    last_val = 0
                    while any(p.is_alive() for p in procs):
                        val = counter.value
                        delta = val - last_val
                        if delta > 0:
                            pbar.update(delta)
                            last_val = val
                        time.sleep(self.pbar_refresh_interval)
                    delta = counter.value - last_val
                    if delta > 0:
                        pbar.update(delta)

                t = threading.Thread(target=_update_pbar, daemon=True)
                t.start()

                # Get output
                ret_list = [None] * self.total
                for _ in range(self.total):
                    idx, result, e = output_queue.get()
                    if e is not None:
                        raise e  # TODO: terminate all processes
                    ret_list[idx] = result

                for p in procs:
                    p.join()
                t.join()

            return ret_list


class AutoBoundedMultiProcess(AutoIterator):
    """
    Auto Bounded Multi-Process

    Create a `BoundedMultiProcess` iterator when `nprocs > 1` and a `SingleProcess` iterator otherwise.

    Parameters:
        func (callable): function to be iterated
        total (int): number of iterations, if not specified, will infer from the list-type `args` and `kwargs` of `run` method
        nprocs (int): number of processes, default is `multiprocessing.cpu_count()`
        use_pool (bool): use `multiprocessing.Pool` or `multiprocessing.Process`, default is True
        pbar_refresh_interval (float): interval of progress bar refreshing when `use_pool` is False, default is 1.0s
        **kwargs: inherited from `AutoIterator`

    Example:
        ```python
        def add(a, b, port):
            print(f"{port}: {a} + {b}")
            return a + b

        a = [1, 2, 3, 4]
        b = [5, 6, 7, 8]
        # Use MultiProcess, length of `port` must be equal to `nprocs`
        res = AutoBoundedMultiProcess(add, nprocs=2).run(a, b, port=[8000, 8001])
        print(res)
        # [6, 8, 10, 12]

        # Use SingleProcess, the type of `port` must be a scalar
        res = AutoBoundedMultiProcess(add, nprocs=1).run(a, b, port=8000)
        print(res)
        # [6, 8, 10, 12]
        ```
    """

    def __init__(
        self,
        func: Callable,
        total: Optional[int] = None,
        nprocs: int = multiprocessing.cpu_count(),
        use_pool: bool = True,
        pbar_refresh_interval: float = 1.0,
        **kwargs,
    ):
        iterator_cls = BoundedMultiProcess if nprocs > 1 else SingleProcess
        super().__init__(
            iterator_cls=iterator_cls,
            func=func,
            total=total,
            nprocs=nprocs,
            use_pool=use_pool,
            pbar_refresh_interval=pbar_refresh_interval,
            **kwargs,
        )


class AsyncProcess(Iterator):
    """
    AsyncProcess

    Parameters:
        func (callable): function to be iterated
        total (int): number of iterations, if not specified, will infer from the list-type `args` and `kwargs` of `run` method
        concurrency (int): concurrent, default is 0
        **kwargs: inherited from `Iterator`

    Example:
        ```python
        async def add(a, b):
            await asyncio.sleep(1)
            return a + b

        a = [1, 2]
        b = [3, 4]
        res = AsyncProcess(add).run(a, b)
        print(res)
        # [4, 6]
        ```
    """

    def __init__(
        self,
        func: Callable,
        total: Optional[int] = None,
        concurrency: int = 0,
        **kwargs,
    ):
        super().__init__(func=func, total=total, **kwargs)

        self.concurrency = concurrency

    async def __call__(self, sem, args):
        idx = args[2]
        if sem is None:
            return idx, await self.partial_func(*args[0], **args[1])
        else:
            async with sem:
                return idx, await self.partial_func(*args[0], **args[1])

    async def main(self):
        if self.concurrency > 0:
            sem = asyncio.Semaphore(self.concurrency)
        else:
            sem = None
        tasks = [
            asyncio.create_task(self(sem, args))
            for args in zip(self.dynamic_args, self.dynamic_kwargs, range(self.total))
        ]
        self.ret_list = [None] * self.total
        for task in tqdm.asyncio.tqdm.as_completed(
            tasks, total=self.total, desc=self.func.__name__, disable=self.disable_pbar
        ):
            idx, result = await task
            self.ret_list[idx] = result

    def run(self, *args, **kwargs):
        """
        run
            - Please ensure that static arguments precede dynamic arguments
            - If the `total` is not specified, it will be inferred from the first list-type parameter
        """
        # parse
        self.parse(*args, **kwargs)

        # run
        asyncio.run(self.main())
        return self.ret_list


class AsyncMultiProcess(Iterator):
    """
    AsyncMultiProcess

    Parameters:
        func (callable): function to be iterated
        total (int): number of iterations, if not specified, will infer from the list-type `args` and `kwargs` of `run` method
        concurrency (int): concurrent, default is 16
        nprocs (int): number of processes, default is `multiprocessing.cpu_count()`
        **kwargs: inherited from `Iterator`

    Example:
        ```python
        async def add(a, b):
            await asyncio.sleep(1)
            return a + b

        a = [1, 2]
        b = [3, 4]
        res = AsyncMultiProcess(add).run(a, b)
        print(res)
        # [4, 6]
        ```
    """

    def __init__(
        self,
        func: Callable,
        total: Optional[int] = None,
        concurrency: int = 16,
        nprocs: int = multiprocessing.cpu_count(),
        **kwargs,
    ):
        super().__init__(func=func, total=total, **kwargs)

        self.concurrency = concurrency
        self.nprocs = nprocs

    async def __call__(self, args):
        return await self.partial_func(*args[0], **args[1])

    async def main(self):
        async with aiomultiprocess.Pool(self.nprocs, childconcurrency=self.concurrency) as p:
            it = p.map(self, zip(self.dynamic_args, self.dynamic_kwargs)).__aiter__()
            self.ret_list = [
                result
                async for result in tqdm.asyncio.tqdm(
                    it, total=self.total, desc=self.func.__name__, disable=self.disable_pbar
                )
            ]

    def run(self, *args, **kwargs):
        """run - Please ensure that static arguments precede dynamic arguments"""

        # parse
        self.parse(*args, **kwargs)

        # run
        asyncio.run(self.main())
        return self.ret_list


class AutoAsyncMultiProcess(AutoIterator):
    """
    Auto AsyncMultiProcess

    Create a `AsyncMultiProcess` iterator when `nprocs > 1` and a `AsyncProcess` iterator otherwise.

    Parameters:
        func (callable): function to be iterated
        total (int): number of iterations, if not specified, will infer from the list-type `args` and `kwargs` of `run` method
        nprocs (int): number of processes, default is `multiprocessing.cpu_count()`
        **kwargs: inherited from `AutoIterator`

    Example:
        ```python
        async def add(a, b):
            await asyncio.sleep(1)
            return a + b

        a = [1, 2]
        b = [3, 4]
        # Use AsyncMultiProcess
        res = AutoAsyncMultiProcess(add, nprocs=2).run(a, b)
        print(res)
        # [4, 6]

        # Use AsyncProcess
        res = AutoAsyncMultiProcess(add, nprocs=1).run(a, b)
        print(res)
        # [4, 6]
        ```
    """

    def __init__(
        self,
        func: Callable,
        total: Optional[int] = None,
        nprocs: int = multiprocessing.cpu_count(),
        **kwargs,
    ):
        iterator_cls = AsyncMultiProcess if nprocs > 1 else AsyncProcess
        super().__init__(
            iterator_cls=iterator_cls,
            func=func,
            total=total,
            nprocs=nprocs,
            **kwargs,
        )


class MultiThread(Iterator):
    """
    MultiThread

    Parameters:
        func (callable): function to be iterated
        total (int): number of iterations, if not specified, will infer from the list-type `args` and `kwargs` of `run` method
        nworkers (int): number of workers, default is 2
        **kwargs: inherited from `Iterator`

    Example:
        ```python
        def add(a, b):
            return a + b

        a = [1, 2]
        b = [3, 4]
        res = MultiThread(add).run(a, b)
        print(res)
        # [4, 6]
        ```
    """

    def __init__(
        self,
        func: Callable,
        total: Optional[int] = None,
        nworkers: int = 2,
        **kwargs,
    ):
        super().__init__(func=func, total=total, **kwargs)

        self.nworkers = nworkers if nworkers > 0 else 2

    def __call__(self, args):
        idx = args[2]
        return idx, self.partial_func(*args[0], **args[1])

    def run(self, *args, **kwargs):
        """
        run
            - Please ensure that static arguments precede dynamic arguments
            - If the `total` is not specified, it will be inferred from the first list-type parameter
        """
        # parse
        self.parse(*args, **kwargs)

        # run
        ret_list = [None] * self.total
        with ThreadPoolExecutor(max_workers=self.nworkers) as p:
            tasks = [p.submit(self, args) for args in zip(self.dynamic_args, self.dynamic_kwargs, range(self.total))]
            for task in tqdm.tqdm(
                as_completed(tasks), total=self.total, desc=self.func.__name__, disable=self.disable_pbar
            ):
                idx, result = task.result()
                ret_list[idx] = result

        return ret_list


class AutoMultiThread(AutoIterator):
    """
    Auto MultiThread

    Create a `MultiThread` iterator when `nworkers > 1` and a `SingleProcess` iterator otherwise.

    Parameters:
        func (callable): function to be iterated
        total (int): number of iterations, if not specified, will infer from the list-type `args` and `kwargs` of `run` method
        nworkers (int): number of workers, default is 2
        **kwargs: inherited from `AutoIterator`

    Example:
        ```python
        def add(a, b):
            return a + b

        a = [1, 2]
        b = [3, 4]
        # Use MultiThread
        res = AutoMultiThread(add, nworkers=2).run(a, b)
        print(res)
        # [4, 6]

        # Use SingleProcess
        res = AutoMultiThread(add, nworkers=1).run(a, b)
        print(res)
        # [4, 6]
        ```
    """

    def __init__(
        self,
        func: Callable,
        total: Optional[int] = None,
        nworkers: int = 2,
        **kwargs,
    ):
        iterator_cls = MultiThread if nworkers > 1 else SingleProcess
        super().__init__(
            iterator_cls=iterator_cls,
            func=func,
            total=total,
            nworkers=nworkers,
            **kwargs,
        )


class ParallelProcess(Iterator):
    """
    ParallelProcess

    A flexible multi-process runner designed to execute a function in parallel,
    supporting both single-task and batch-task execution. It automatically splits
    input arguments across processes, handles static vs dynamic arguments, and
    provides optional progress bar support via shared counters.

    Suitable for cases where:
    - Each process handles a batch of tasks (is_single_task_func=False)
    - Each task is processed independently (is_single_task_func=True)

    Parameters:
        func (callable): function to be iterated
        total (int | None): number of iterations, if not specified, will infer from the list-type `args` and `kwargs` of `run` method
        nprocs (int): number of processes, default is `multiprocessing.cpu_count()`
        is_single_task_func (bool): whether the function is single task, default is True (when version >= 1.11.1)
        pbar_refresh_interval (float): interval of progress bar refreshing, default is 1.0s
        flatten_result (bool): whether to flatten the result for multi-task function, default is True
        **kwargs: inherited from `Iterator`

    Example:
        ```python
        # For multi-task function
        def add(arr1, arr2):
            return [a + b for a, b in zip(arr1, arr2)]

        a = [1, 2, 3, 4]
        b = [3, 4, 5, 6]
        res = ParallelProcess(add, nprocs=2, is_single_task_func=False).run(a, b)
        print(res)
        # [[4, 6], [8, 10]]

        # use `_counter` argument to display progress bar
        def add(arr1, arr2, _counter=None):
            res = []
            for a, b in zip(arr1, arr2):
                res.append(a + b)
                with _counter.get_lock():
                    _counter.value += 1
            return res

        a = [1, 2, 3, 4]
        b = [3, 4, 5, 6]
        res = ParallelProcess(add, nprocs=2, is_single_task_func=False).run(a, b)
        print(res)
        # [[4, 6], [8, 10]]

        # For single-task function
        def add(a, b):
            return a + b

        a = [1, 2, 3, 4]
        b = [3, 4, 5, 6]
        res = ParallelProcess(add, nprocs=2, is_single_task_func=True).run(a, b)
        print(res)
        # [4, 6, 8, 10]
        ```
    """

    def __init__(
        self,
        func: Callable,
        total: Optional[int] = None,
        nprocs: int = multiprocessing.cpu_count(),
        is_single_task_func: bool = True,
        pbar_refresh_interval: float = 1.0,
        **kwargs,
    ):
        if is_single_task_func:
            func = SingleProcess(func).run
        super().__init__(func=func, total=total, **kwargs)

        self.nprocs = nprocs if nprocs > 0 else multiprocessing.cpu_count()
        self.is_single_task_func = is_single_task_func
        self.pbar_refresh_interval = pbar_refresh_interval

    def parse(self, *args, **kwargs):
        """parse"""
        # infer the number of expected iterations from args
        if self.total is None:
            for arg in args:
                if isinstance(arg, list):
                    self.total = len(arg)
                    break

        # infer the number of expected iterations from kwargs
        if self.total is None:
            for _, value in kwargs.items():
                if isinstance(value, list):
                    self.total = len(value)
                    break

        assert self.total is not None, "Can not infer the number of expected iterations"
        if self.total == self.nprocs:
            print(
                "Warning: When resource arguments are provided, please ensure that `total != nprocs`; otherwise, it will be ambiguous whether the argument represents a resource or not"
            )

        # split tasks
        indices = np.linspace(0, self.total, self.nprocs + 1).astype("int32")

        # parse args
        self.dynamic_args = [[] for _ in range(self.nprocs)]
        self.static_args = []
        dynamic_arg_flags = []
        for arg in args:
            if isinstance(arg, list):
                assert (
                    len(arg) == self.total or len(arg) == self.nprocs
                ), f"{len(arg)} != {self.total} and {len(arg)} != {self.nprocs}"
                if len(arg) == self.total:
                    for i in range(self.nprocs):
                        self.dynamic_args[i].append(arg[indices[i] : indices[i + 1]])
                else:
                    for i in range(self.nprocs):
                        self.dynamic_args[i].append(arg[i])
                dynamic_arg_flags.append(True)
            else:
                self.static_args.append(arg)
                dynamic_arg_flags.append(False)

        # Assert static arguments precede dynamic arguments
        if len(dynamic_arg_flags) > 1:
            dynamic_arg_flags = np.array(dynamic_arg_flags)
            if (dynamic_arg_flags[:-1] > dynamic_arg_flags[1:]).any():
                msg = ["dynamic" if flag else "static" for flag in dynamic_arg_flags]
                raise RuntimeError(f"Dynamic arguments precede static arguments: {msg}")

        # parse kwargs
        self.dynamic_kwargs = [{} for _ in range(self.nprocs)]
        self.static_kwargs = {}
        for key, value in kwargs.items():
            if isinstance(value, list):
                assert (
                    len(value) == self.total or len(value) == self.nprocs
                ), f"{len(value)} != {self.total} or {len(value)} != {self.nprocs}"
                if len(value) == self.total:
                    for i in range(self.nprocs):
                        self.dynamic_kwargs[i][key] = value[indices[i] : indices[i + 1]]
                else:
                    for i in range(self.nprocs):
                        self.dynamic_kwargs[i][key] = value[i]
            else:
                self.static_kwargs[key] = value

        # fix the static args
        self.partial_func = functools.partial(self.func, *self.static_args, **self.static_kwargs)

    def __call__(self, dynamic_args, dynamic_kwargs, results_dict, idx):
        res = self.partial_func(*dynamic_args, **dynamic_kwargs)
        results_dict[idx] = res

    def run(self, *args, **kwargs):
        """
        run
            - Please ensure that static arguments precede dynamic arguments
            - If the `total` is not specified, it will be inferred from the first list-type parameter
        """
        # parse
        self.parse(*args, **kwargs)

        # run
        procs = []
        manager = multiprocessing.Manager()
        results_dict = manager.dict()

        # check if function has `_counter` args
        has_counter = False
        if not self.is_single_task_func:
            params = inspect.signature(self.func).parameters
            has_counter = "_counter" in params

        if self.is_single_task_func or has_counter:
            counter = multiprocessing.Value("i", 0)
            for i in range(self.nprocs):
                p = multiprocessing.Process(
                    target=self,
                    args=(
                        self.dynamic_args[i],
                        {**self.dynamic_kwargs[i], "_counter": counter},
                        results_dict,
                        i,
                    ),
                )
                procs.append(p)
                p.start()

            # display progress bar
            with tqdm.tqdm(total=self.total, desc=self.func.__name__, disable=self.disable_pbar) as pbar:

                def _update_pbar():
                    """update function"""
                    last_val = 0
                    while any(p.is_alive() for p in procs):
                        val = counter.value
                        delta = val - last_val
                        if delta > 0:
                            pbar.update(delta)
                            last_val = val
                        time.sleep(self.pbar_refresh_interval)
                    delta = counter.value - last_val
                    if delta > 0:
                        pbar.update(delta)

                t = threading.Thread(target=_update_pbar, daemon=True)
                t.start()

                for p in procs:
                    p.join()
                t.join()

            ret_list = [results_dict[idx] for idx in range(self.nprocs)]
        else:
            for i in range(self.nprocs):
                p = multiprocessing.Process(
                    target=self,
                    args=(
                        self.dynamic_args[i],
                        self.dynamic_kwargs[i],
                        results_dict,
                        i,
                    ),
                )
                procs.append(p)
                p.start()

            for p in procs:
                p.join()

            ret_list = [results_dict[idx] for idx in range(self.nprocs)]

        # flatten the ret_list from 2d to 1d
        ret_list = [_ for sub_ret_list in ret_list for _ in sub_ret_list]

        return ret_list


class AutoParallelProcess(AutoIterator):
    """
    Auto ParallelProcess

    Create a `ParallelProcess` iterator when `nworkers > 1` and a `SingleProcess` iterator otherwise.

    Parameters:
        func (callable): function to be iterated
        total (int): number of iterations, if not specified, will infer from the list-type `args` and `kwargs` of `run` method
        nprocs (int): number of processes, default is `multiprocessing.cpu_count()`
        is_single_task_func (bool): whether the function is single task, default is True (when version >= 1.11.1)
        **kwargs: inherited from `AutoIterator`

    Example:
        ```python
        # For multi-task function
        def add(arr1, arr2):
            return [a + b for a, b in zip(arr1, arr2)]

        a = [1, 2, 3, 4]
        b = [3, 4, 5, 6]
        # Use ParallelProcess
        res = AutoParallelProcess(add, nprocs=2, is_single_task_func=False).run(a, b)
        print(res)
        # [[4, 6], [8, 10]]

        # Use original function
        res = AutoParallelProcess(add, nprocs=1, is_single_task_func=False).run(a, b)
        print(res)
        # [[4, 6], [8, 10]]

        # For single-task function
        def add(a, b):
            return a + b

        a = [1, 2, 3, 4]
        b = [3, 4, 5, 6]
        # Use ParallelProcess
        res = AutoParallelProcess(add, nprocs=2, is_single_task_func=True).run(a, b)
        print(res)
        # [4, 6, 8, 10]

        # Use SingleProcess
        res = AutoParallelProcess(add, nprocs=1, is_single_task_func=True).run(a, b)
        print(res)
        # [4, 6, 8, 10]
        ```
    """

    def __init__(
        self,
        func: Callable,
        total: Optional[int] = None,
        nprocs: int = multiprocessing.cpu_count(),
        is_single_task_func: bool = True,
        **kwargs,
    ):
        iterator_cls = ParallelProcess if nprocs > 1 else (SingleProcess if is_single_task_func else None)
        super().__init__(
            iterator_cls=iterator_cls,
            func=func,
            total=total,
            nprocs=nprocs,
            is_single_task_func=is_single_task_func,
            **kwargs,
        )



import os
import glob
import imageio.v2 as iio
import argparse
from lh_tool.iterator import SingleProcess, MultiProcess


def image2gif(image_path, gif_file, postfix, duration, speed=1):
    image_file_list = sorted(glob.glob(os.path.join(image_path, f"*.{postfix}")))
    if len(image_file_list) == 0:
        return
    gif_file = os.path.abspath(image_path) + ".gif" if gif_file is None else gif_file

    image_file_list = [image_file_list[i] for i in range(0, len(image_file_list), speed)]
    images = [iio.imread(image_file) for image_file in image_file_list]
    duration /= len(image_file_list) - 1
    iio.mimsave(gif_file, images, duration=duration)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", type=str, default=".", help="path of image files")
    parser.add_argument("-o", "--output", type=str, help="output gif file")
    parser.add_argument(
        "-p",
        "--postfix",
        type=str,
        default="png",
        help="postfix of image filename",
    )
    parser.add_argument("-d", "--duration", type=float, default=1, help="duration")
    parser.add_argument("-a", "--speed", type=int, default=1, help="speed for video")
    parser.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="convert images to gif recursively",
    )
    parser.add_argument("-n", "--nprocs", type=int, default=1, help="number of process")
    opts = parser.parse_args()
    print(opts)

    try:
        image_path = opts.input
        gif_file = opts.output
        postfix = opts.postfix
        duration = opts.duration
        speed = opts.speed
        recursive = opts.recursive
        nprocs = opts.nprocs
        if recursive:
            image_path_list = glob.glob(os.path.join(opts.input, "*/"))
            if nprocs == 1:
                iterator = SingleProcess(image2gif)
            else:
                iterator = MultiProcess(image2gif, nprocs=nprocs)
            iterator.run(
                image_path_list,
                gif_file=None,
                postfix=postfix,
                duration=duration,
                speed=speed,
            )
        else:
            image2gif(image_path, gif_file, postfix, duration, speed)

    except AssertionError as e:
        print(e)


if __name__ == "__main__":
    main()



import os
import cv2
import glob
import numpy as np
import argparse
from lh_tool.iterator import SingleProcess, MultiProcess
import imageio.v2 as iio


def concat_image(image_file_list, output_file, padding_size=5, hstack=True, black=True):
    assert len(image_file_list), "The number of image files must greater than 0"

    color = 0 if black else 255
    image = None
    height, width = None, None
    padding = None
    for image_file in image_file_list:
        input_image = iio.imread(image_file)
        h, w = input_image.shape[:2]
        if image is None:
            image = input_image
            height, width = h, w
            padding_shape = list(input_image.shape)
            if hstack:
                padding_shape[1] = padding_size
            else:
                padding_shape[0] = padding_size
            padding = np.full(padding_shape, color, dtype="uint8") * color
        else:
            if hstack:
                input_image = cv2.resize(input_image, (round(w * height / h), height))
                image = np.hstack((image, padding, input_image))
            else:
                input_image = cv2.resize(input_image, (width, round(h * width / w)))
                image = np.vstack((image, padding, input_image))

    iio.imwrite(output_file, image)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        default=".",
        help="image file or path of input image files",
    )
    parser.add_argument(
        "-a",
        "--another_input",
        type=str,
        help="another image file or path of input image files",
    )
    parser.add_argument("-o", "--output", type=str, help="output image file or output path")
    parser.add_argument(
        "-p",
        "--postfix",
        type=str,
        default="png",
        help="postfix of image filename",
    )
    parser.add_argument("-c", "--cover", action="store_true", help="cover the input image file")
    parser.add_argument("-s", "--padding_size", type=int, default=5, help="padding size")
    parser.add_argument(
        "-v",
        "--vertical",
        dest="hstack",
        action="store_false",
        help="vertical stack",
    )
    parser.add_argument(
        "-w",
        "--white",
        dest="black",
        action="store_false",
        help="padding with white",
    )
    parser.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="concatenate images recursively",
    )
    parser.add_argument("-n", "--nprocs", type=int, default=1, help="number of process")
    opts = parser.parse_args()
    print(opts)

    try:
        postfix = opts.postfix
        cover = opts.cover
        padding_size = opts.padding_size
        hstack = opts.hstack
        black = opts.black
        recursive = opts.recursive
        nprocs = opts.nprocs
        if recursive:
            image_file_list = sorted(glob.glob(os.path.join(opts.input, f"*.{postfix}")))
            another_image_file_list = sorted(glob.glob(os.path.join(opts.another_input, f"*.{postfix}")))
            assert len(image_file_list) == len(another_image_file_list), (
                "The number of image files under the two " "paths is not equal"
            )
            output_path = opts.output
            assert cover or output_path is not None, "Output path is 'None'"
            output_path = opts.input if cover else output_path
            os.makedirs(output_path, exist_ok=True)
            output_file_list = [
                os.path.join(output_path, os.path.basename(image_file)) for image_file in image_file_list
            ]
            if nprocs == 1:
                iterator = SingleProcess(concat_image)
            else:
                iterator = MultiProcess(concat_image, nprocs=nprocs)
            iterator.run(
                list(zip(image_file_list, another_image_file_list)),
                output_file_list,
                padding_size=padding_size,
                hstack=hstack,
                black=black,
            )

        else:
            if os.path.isdir(opts.input):
                image_file_list = sorted(glob.glob(os.path.join(opts.input, f"*.{postfix}")))
                output_file = os.path.abspath(opts.input) + f".{postfix}" if opts.output is None else opts.output
                concat_image(image_file_list, output_file, padding_size, hstack, black)
            else:
                image_file = opts.input
                another_image_file = opts.another_input
                image_file_list = [image_file, another_image_file]
                output_file = opts.output
                assert cover or output_file is not None, "Output file is 'None'"

                output_file = image_file if cover else output_file
                concat_image(image_file_list, output_file, padding_size, hstack, black)

    except AssertionError as e:
        print(e)


if __name__ == "__main__":
    main()



import os
import cv2
import glob
import tqdm
import argparse
from lh_tool.iterator import SingleProcess, MultiProcess


def images2video(image_path, video_file, postfix, fourcc, fps, frameSize=None, speed=1):
    image_file_list = sorted(glob.glob(os.path.join(image_path, f"*.{postfix}")))
    if len(image_file_list) == 0:
        return
    video_file = os.path.abspath(image_path) + ".mp4" if video_file is None else video_file

    if frameSize is None:
        image = cv2.imread(image_file_list[0])
        image = image[:, :, ::-1]  # RGB2BGR
        frameSize = (image.shape[1], image.shape[0])
    videoWriter = cv2.VideoWriter(video_file, fourcc, fps, frameSize)
    assert videoWriter.isOpened(), f"Failed to create file: {video_file}"

    image_file_list = [image_file_list[i] for i in range(0, len(image_file_list), speed)]
    for image_file in tqdm.tqdm(image_file_list, desc=video_file):
        image = cv2.imread(image_file)
        if frameSize != (image.shape[1], image.shape[0]):
            image = cv2.resize(image, tuple(frameSize))
        videoWriter.write(image)
    videoWriter.release()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", type=str, default=".", help="path of image files")
    parser.add_argument("-o", "--output", type=str, help="output video file")
    parser.add_argument(
        "-p",
        "--postfix",
        type=str,
        default="png",
        help="postfix of image filename",
    )
    parser.add_argument("-f", "--fps", type=float, default=29.97, help="desired fps for video")
    parser.add_argument("-s", "--size", type=int, nargs=2, help="desired frame size for video")
    parser.add_argument("-a", "--speed", type=int, default=1, help="speed for video")
    parser.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="convert video to images recursively",
    )
    parser.add_argument("-n", "--nprocs", type=int, default=1, help="number of process")
    opts = parser.parse_args()
    print(opts)

    try:
        image_path = opts.input
        video_file = opts.output
        postfix = opts.postfix
        fps = opts.fps
        size = None if opts.size is None else tuple(opts.size)
        speed = opts.speed
        recursive = opts.recursive
        nprocs = opts.nprocs
        fourcc = cv2.VideoWriter.fourcc("m", "p", "4", "v")
        if recursive:
            image_path_list = glob.glob(os.path.join(opts.input, "*/"))
            if nprocs == 1:
                iterator = SingleProcess(images2video)
            else:
                iterator = MultiProcess(images2video, nprocs=nprocs)
            iterator.run(
                image_path_list,
                video_file=None,
                postfix=postfix,
                fourcc=fourcc,
                fps=fps,
                frameSize=size,
                speed=speed,
            )
        else:
            images2video(image_path, video_file, postfix, fourcc, fps, size, speed)

    except AssertionError as e:
        print(e)


if __name__ == "__main__":
    main()



# -*- encoding: utf-8 -*-
"""
@File    :   time_consumption.py
@Time    :   2024/03/14 10:53:51
@Author  :   lihao57
@Version :   1.0
@Contact :   lihao57@baidu.com
"""


import time
from typing import Callable, Optional
import functools
import asyncio


class TimeConsumptionDecorator:
    """
    TimeConsumptionDecorator used to print the module time consumption (Deprecated)

    Parameters:
        print_func (callable): The print function used to print the time consumption, default is `print`

    Example:
        ```python
        @TimeConsumptionDecorator()
        def test1():
            time.sleep(1)

        test1()

        # using logging to print
        import logging
        @TimeConsumptionDecorator(logging.info)
        def test2():
            time.sleep(1)

        test2()
        ```
    """

    def __init__(self, print_func: Callable = print):
        self._print_func = print_func

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            """wrapper"""
            start_time = time.perf_counter()
            ret = func(*args, **kwargs)
            end_time = time.perf_counter()
            self._print_func("{} time consuming: {}".format(str(func).split(" ")[1], end_time - start_time))
            return ret

        wrapper.__name__ = func.__name__
        return wrapper


class TimeConsumptionContextManager:
    """
    TimeConsumptionContextManager used to print the module time consumption (Deprecated)

    Parameters:
        context (str): The context used to print the time consumption, default is ''
        print_func (callable): The print function used to print the time consumption, default is `print`

    Example:
        ```python
        import logging
        with TimeConsumptionContextManager('block1'):
            time.sleep(1)

        # using logging to print
        with TimeConsumptionContextManager('block2', logging.error):
            time.sleep(1)
        ```
    """

    def __init__(self, context: str = "", print_func: Callable = print):
        self._context = context
        self._print_func = print_func

    def __enter__(self):
        self.start_time = time.perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end_time = time.perf_counter()
        if self._context:
            self._print_func("{} time consuming: {}".format(self._context, self.end_time - self.start_time))
        else:
            self._print_func("time consuming: {}".format(self.end_time - self.start_time))


class TimeConsumption:
    """
    TimeConsumption used to print the module time consumption

    Parameters:
        context (str): The context used to print the time consumption, default is ''
        print_func (callable): The print function used to print the time consumption, default is `print`
        format_func (Optional[Callable]): The format function used to format the output string, default is None
        sync (bool): Whether to synchronize the cuda before and after the execution of the code block, default is False

    Example:
        ```python
        # used as decorator
        @TimeConsumption()
        def func():
            time.sleep(1)

        func()

        # used as context manager
        with TimeConsumption("block"):
            time.sleep(1)

        # for pytorch code block, you need to set `sync=True`
        with TimeConsumption("block", sync=True):
            x = torch.randn((1000, 1000)).to("cuda")
            y = torch.randn((1000, 1000)).to("cuda")
            z = x + y

        # for async function
        import asyncio

        @TimeConsumption()
        async def func():
            await asyncio.sleep(0.1)

        asyncio.run(func())

        async def func():
            async with TimeConsumption():
                await asyncio.sleep(0.1)
        asyncio.run(func())
        ```
    """

    def __init__(
        self,
        context: str = "",
        print_func: Callable = print,
        format_func: Optional[Callable] = None,
        sync: bool = False,
    ):
        self._context = context
        self._print_func = print_func
        self._format_func = format_func
        self._sync = sync
        if self._sync:
            from torch.cuda import synchronize

            self._synchronize = synchronize
        else:
            self._synchronize = lambda: None

    def __call__(self, func):
        if asyncio.iscoroutinefunction(func):

            @functools.wraps(func)
            async def wrapper(*args, **kwargs):
                """wrapper"""
                start_time = time.perf_counter()
                self._synchronize()
                ret = await func(*args, **kwargs)
                self._synchronize()
                delta_time = time.perf_counter() - start_time
                if self._format_func is not None:
                    delta_time = self._format_func(delta_time)
                context = self._context if self._context else str(func).split(" ")[1]
                self._print_func(f"{context} time consuming: {delta_time}")
                return ret

        else:

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                """wrapper"""
                start_time = time.perf_counter()
                self._synchronize()
                ret = func(*args, **kwargs)
                self._synchronize()
                delta_time = time.perf_counter() - start_time
                if self._format_func is not None:
                    delta_time = self._format_func(delta_time)
                context = self._context if self._context else str(func).split(" ")[1]
                self._print_func(f"{context} time consuming: {delta_time}")
                return ret

        wrapper.__name__ = func.__name__
        return wrapper

    def __enter__(self):
        self.start_time = time.perf_counter()
        self._synchronize()
        return self

    async def __aenter__(self):
        return self.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._synchronize()
        delta_time = time.perf_counter() - self.start_time
        if self._format_func is not None:
            delta_time = self._format_func(delta_time)
        if self._context:
            self._print_func(f"{self._context} time consuming: {delta_time}")
        else:
            self._print_func(f"time consuming: {delta_time}")

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return self.__exit__(exc_type, exc_val, exc_tb)


def time_consumption(
    context: str = "",
    print_func: Callable = print,
    format_func: Optional[Callable] = None,
    sync: bool = False,
):
    """
    `time_consumption` used to print the module time consumption

    Parameters:
        context (str): The context used to print the time consumption, default is ''
        print_func (callable): The print function used to print the time consumption, default is `print`
        format_func (Optional[Callable]): The format function used to format the output string, default is None
        sync (bool): Whether to synchronize the cuda before and after the execution of the code block, default is False
    Example:
        ```python
        # used as decorator
        @time_consumption()
        def test():
            time.sleep(1)

        test()

        # used as context manager
        with time_consumption("block"):
            time.sleep(1)

        # for pytorch code block, you need to set `sync=True`
        with time_consumption("block", sync=True):
            x = torch.randn((1000, 1000)).to("cuda")
            y = torch.randn((1000, 1000)).to("cuda")
            z = x + y

        # for async function
        import asyncio

        @time_consumption()
        async def func():
            await asyncio.sleep(0.1)

        asyncio.run(func())

        async def func():
            async with time_consumption():
                await asyncio.sleep(0.1)
        asyncio.run(func())
        ```
    """
    return TimeConsumption(context=context, print_func=print_func, format_func=format_func, sync=sync)



# -*- encoding: utf-8 -*-
"""
@File    :   email.py
@Time    :   2024/03/14 10:59:46
@Author  :   lihao57
@Version :   1.0
@Contact :   lihao57@baidu.com
"""

import os
import smtplib
from yacs.config import CfgNode
from email import encoders
from email.header import Header
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email.utils import formataddr


class Email:
    """
    Email

    Parameters:
        cfg (dict): config

    Example:
        ```python
        Email(cfg).send()
        ```
    """

    def __init__(self, cfg):
        self.from_name = cfg.from_name
        self.from_addr = cfg.from_addr
        self.password = cfg.password
        self.to_name = cfg.to_name
        self.to_addr = cfg.to_addr
        self.subject = cfg.subject
        self.content = cfg.content
        self.smtp_server = cfg.smtp_server
        self.port = cfg.port
        self.encoding = cfg.encoding
        self.file_list = cfg.file_list

    def send(self):
        # 配置基础信息
        msg = MIMEMultipart()
        msg["From"] = formataddr(
            (
                Header(self.from_name, charset=self.encoding).encode(),
                self.from_addr,
            )
        )
        msg["To"] = formataddr(
            (
                Header(self.to_name, charset=self.encoding).encode(),
                self.to_addr,
            )
        )
        msg["Subject"] = Header(self.subject, charset=self.encoding).encode()
        msg.attach(MIMEText(self.content, _charset=self.encoding))

        # 添加附件
        if self.file_list is not None:
            for filename in self.file_list:
                with open(filename, "rb") as f:
                    _, extension = os.path.splitext(filename)
                    mime = MIMEBase(extension, extension, filename=filename)
                    mime.add_header("Content-Disposition", "attachment", filename=filename)
                    mime.add_header("Content-ID", "<0>")
                    mime.add_header("X-Attachment-Id", "0")
                    mime.set_payload(f.read())
                    encoders.encode_base64(mime)
                    msg.attach(mime)

        # 发送邮件
        server = smtplib.SMTP(self.smtp_server, self.port)
        server.login(self.from_addr, self.password)
        server.sendmail(self.from_addr, [self.to_addr], msg.as_string())
        server.quit()
        print("发送成功")


class CustomEmail(Email):
    """
    CustomEmail

    Parameters:
        from_name (str): from name
        from_addr (str): from address
        password (str): password
        to_name (str): to name
        to_addr (str): to address
        subject (str): subject
        content (str): content

    Example:
        ```python
        CustomEmail(from_name, from_addr, password, to_name, to_addr, subject, content).send()
        ```
    """

    def __init__(
        self,
        from_name,
        from_addr,
        password,
        to_name,
        to_addr,
        subject="",
        content="",
    ):
        cfg = CfgNode()
        cfg.from_name = from_name
        cfg.from_addr = from_addr
        cfg.password = password
        cfg.to_name = to_name
        cfg.to_addr = to_addr
        cfg.subject = subject
        cfg.content = content
        cfg.smtp_server = "smtp.whu.edu.cn"
        cfg.port = 25
        cfg.encoding = "utf-8"
        cfg.file_list = None
        cfg.freeze()

        super().__init__(cfg)



import os
import shutil
import argparse
import glob
from lh_tool.image2pdf import image2pdf
from lh_tool.pdf2image import pdf2image


def compress_pdf(input_file, output_file, temp_path, zoom):
    temp_path = os.path.splitext(input_file)[0] if temp_path is None else temp_path
    assert not os.path.exists(temp_path), f"'{temp_path}' exists"

    pdf2image(input_file, None, temp_path, zoom)
    image_file_list = sorted(glob.glob(os.path.join(temp_path, "*.png")))
    image2pdf(image_file_list, output_file)
    shutil.rmtree(temp_path)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", type=str, help="input pdf file", required=True)
    parser.add_argument("-o", "--output", type=str, help="output pdf file", required=True)
    parser.add_argument("-t", "--temp_path", type=str, help="temporary path for image files")
    parser.add_argument("-z", "--zoom", type=float, default=1, help="zoom scale")
    opts = parser.parse_args()
    print(opts)

    try:
        input_file = opts.input
        output_file = opts.output
        temp_path = opts.temp_path
        zoom = opts.zoom
        compress_pdf(input_file, output_file, temp_path, zoom)
    except AssertionError as e:
        print(e)


if __name__ == "__main__":
    main()



import os
import argparse
import fitz
import glob


def image2pdf(image_file_list, pdf_file):
    if len(image_file_list) == 0:
        return

    doc = fitz.open()
    for image_file in image_file_list:
        imgdoc = fitz.open(image_file)
        pdfbytes = imgdoc.convert_to_pdf()
        imgpdf = fitz.open("pdf", pdfbytes)
        doc.insert_pdf(imgpdf)
    doc.save(pdf_file)
    doc.close()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        default=".",
        help="input image file or input path",
    )
    parser.add_argument("-o", "--output", type=str, help="output pdf file")
    parser.add_argument(
        "-p",
        "--postfix",
        type=str,
        default="png",
        help="postfix of image filename valid only if " "'input' is a path)",
    )
    opts = parser.parse_args()
    print(opts)

    if os.path.isdir(opts.input):
        image_file_list = sorted(glob.glob(os.path.join(opts.input, f"*.{opts.postfix}")))
        pdf_file = os.path.abspath(opts.input) + ".pdf" if opts.output is None else opts.output
    else:
        image_file_list = [opts.input]
        pdf_file = os.path.splitext(opts.input)[0] + ".pdf" if opts.output is None else opts.output

    image2pdf(image_file_list, pdf_file)


if __name__ == "__main__":
    main()



# -*- encoding: utf-8 -*-
"""
@File    :   test_time_consumption.py
@Time    :   2024/03/14 13:46:12
@Author  :   lihao57
@Version :   1.0
@Contact :   lihao57@baidu.com
"""


import unittest
import time
import sys
import asyncio

sys.path.append("../src")
from lh_tool.time_consumption import (
    TimeConsumptionDecorator,
    TimeConsumptionContextManager,
    TimeConsumption,
    time_consumption,
)


class TestConsumption(unittest.TestCase):
    """TestConsumption"""

    @staticmethod
    def callback(times):
        """callback"""
        times.append(time.time())

    def test_time_consumption_decorator(self):
        """test_time_consumption_decorator"""

        @TimeConsumptionDecorator()
        def func():
            """func"""
            time.sleep(0.1)

        func()
        self.assertTrue(True)

    def test_time_consumption_context_manager(self):
        """test time consumption context manager"""
        with TimeConsumptionContextManager():
            time.sleep(0.1)
        self.assertTrue(True)

    def test_time_consumption_class(self):
        """test time consumption class"""

        @TimeConsumption()
        def func():
            """func"""
            time.sleep(0.1)

        func()

        with TimeConsumption():
            time.sleep(0.1)
        self.assertTrue(True)

    def test_async_time_consumption_class(self):
        """test async time consumption class"""

        @TimeConsumption()
        async def func():
            """func"""
            await asyncio.sleep(0.1)

        asyncio.run(func())

        async def func():
            async with TimeConsumption():
                await asyncio.sleep(0.1)

        asyncio.run(func())
        self.assertTrue(True)

    def test_time_consumption_function(self):
        """test time consumption function"""

        @time_consumption()
        def func():
            """func"""
            time.sleep(0.1)

        func()

        with time_consumption():
            time.sleep(0.1)
        self.assertTrue(True)


if __name__ == "__main__":
    unittest.main()



# -*- encoding: utf-8 -*-
"""
@File    :   rename_file.py
@Time    :   2024/03/22 12:40:50
@Author  :   lihao57
@Version :   1.0
@Contact :   lihao57@baidu.com
"""


import os
import glob
import argparse
from lh_tool.iterator import SingleProcess, MultiProcess


def rename(input_file, output_file):
    """rename file"""
    if input_file != output_file:
        os.rename(input_file, output_file)


def main():
    example = """
        example:
            # rename filename
            rename_file -i input -e "lambda x: '{:06d}.{}'.format(int(x.split('.')[0]), x.split('.')[1])"

            # rename file extension
            rename_file -i input -e "lambda x: x.replace('.png', '.jpg')"
    """
    parser = argparse.ArgumentParser(usage=example)
    parser.add_argument("-i", "--input", type=str, default=".", help="path of input files")
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="path of output files. The default is the same as the " "path of input files",
    )
    parser.add_argument(
        "-e",
        "--expression",
        type=str,
        required=True,
        help="rename lambda expression",
    )
    parser.add_argument("-r", "--recursive", action="store_true", help="rename recursively")
    parser.add_argument("-n", "--nprocs", type=int, default=1, help="number of process")

    opts = parser.parse_args()
    print(opts)

    try:
        input_path = opts.input
        output_path = opts.output
        expression = opts.expression
        expression_func = eval(expression)
        recursive = opts.recursive
        nprocs = opts.nprocs
        if recursive:
            input_file_list = glob.glob(
                os.path.join(input_path, "**", "*"),
                recursive=True,
            )
        else:
            input_file_list = glob.glob(os.path.join(input_path, "*"))
        output_file_list = []
        for input_file in input_file_list:
            filename = os.path.basename(input_file)
            filename = expression_func(filename)
            if output_path is not None:
                output_file = os.path.join(output_path, filename)
            else:
                dirname = os.path.dirname(input_file)
                output_file = os.path.join(dirname, filename)
            output_file_list.append(output_file)

        assert len(output_file_list) == len(
            set(output_file_list)
        ), "output file list is not unique, please check your expression: {}".format(expression)

        process = MultiProcess if nprocs > 1 else SingleProcess
        process(rename, nprocs=nprocs).run(input_file_list, output_file_list)

    except AssertionError as e:
        print(e)


if __name__ == "__main__":
    main()



# -*- encoding: utf-8 -*-
"""
@File    :   test_iterator.py
@Time    :   2024/03/14 13:29:51
@Author  :   lihao57
@Version :   1.0
@Contact :   lihao57@baidu.com
"""


import unittest
import asyncio
import sys
import time

sys.path.append("../src")
from lh_tool.iterator import (
    SingleProcess,
    MultiProcess,
    AutoMultiProcess,
    BoundedMultiProcess,
    AutoBoundedMultiProcess,
    MultiThread,
    AutoMultiThread,
    AsyncProcess,
    AsyncMultiProcess,
    AutoAsyncMultiProcess,
    ParallelProcess,
    AutoParallelProcess,
)


class TestIterator(unittest.TestCase):
    """Test Iterator"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.a = [i for i in range(10)]
        self.b = [i for i in range(10)]
        self.t = [5 if i == 0 else 1 for i in range(10)]
        self.res = [i + j for i, j in zip(self.a, self.b)]
        self.disable_pbar = False

    @staticmethod
    def process(a, b, t, opt="+"):
        """process"""
        time.sleep(t)
        if opt == "+":
            return a + b
        else:
            return a - b

    @staticmethod
    def bounded_process(a, b, t, opt="+", port=8000):
        """resource slot process"""
        time.sleep(t)
        print(f"{port}: {a} {opt} {b}")
        if opt == "+":
            return a + b
        else:
            return a - b

    @staticmethod
    def process_batch(a_list, b_list, t_list, opt="+"):
        """process batch"""
        res_list = []
        for a, b, t in zip(a_list, b_list, t_list):
            time.sleep(t)
            if opt == "+":
                res = a + b
            else:
                res = a - b
            res_list.append(res)
        return res_list

    @staticmethod
    def process_batch_v2(a_list, b_list, t_list, opt="+", _counter=None):
        """process batch"""
        res_list = []
        for a, b, t in zip(a_list, b_list, t_list):
            time.sleep(t)
            if opt == "+":
                res = a + b
            else:
                res = a - b
            res_list.append(res)
            if _counter is not None:
                with _counter.get_lock():
                    _counter.value += 1
        return res_list

    @staticmethod
    async def async_process(a, b, t, opt="+"):
        """async process"""
        await asyncio.sleep(t)
        if opt == "+":
            return a + b
        else:
            return a - b

    def test_single_process(self):
        """test single process"""
        ret_list = SingleProcess(self.process, disable_pbar=self.disable_pbar).run(self.a, self.b, self.t, opt="+")
        self.assertEqual(ret_list, self.res)
        exception = None
        try:
            ret_list = SingleProcess(self.process, disable_pbar=self.disable_pbar).run(self.a, self.b, self.t, "+")
        except Exception as e:
            exception = e
            print(e)
        self.assertIsInstance(exception, RuntimeError)

    def test_multi_process(self):
        """test multi process"""
        ret_list = MultiProcess(self.process, nprocs=10, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+"
        )
        self.assertEqual(ret_list, self.res)
        ret_list = MultiProcess(self.process, nprocs=10, use_pool=False, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+"
        )
        self.assertEqual(ret_list, self.res)

    def test_auto_multi_process(self):
        """test auto multi process"""
        ret_list = AutoMultiProcess(self.process, nprocs=10, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+"
        )
        self.assertEqual(ret_list, self.res)
        ret_list = AutoMultiProcess(self.process, nprocs=10, use_pool=False, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+"
        )
        self.assertEqual(ret_list, self.res)
        ret_list = AutoMultiProcess(self.process, nprocs=1, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+"
        )
        self.assertEqual(ret_list, self.res)

    def test_bounded_multi_process(self):
        """test resource slot multi process"""
        ret_list = BoundedMultiProcess(self.bounded_process, nprocs=2, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+", port=[8000, 8001]
        )
        self.assertEqual(ret_list, self.res)
        ret_list = BoundedMultiProcess(
            self.bounded_process, nprocs=2, use_pool=False, disable_pbar=self.disable_pbar
        ).run(self.a, self.b, self.t, opt="+", port=[8000, 8001])
        self.assertEqual(ret_list, self.res)

        exception = None
        try:
            ret_list = BoundedMultiProcess(
                self.bounded_process, nprocs=len(self.a), disable_pbar=self.disable_pbar
            ).run(self.a, self.b, self.t, opt="+", port=list(range(8000, 8000 + len(self.a))))
        except Exception as e:
            exception = e
            print(e)
        self.assertIsInstance(exception, AssertionError)

    def test_auto_bounded_multi_process(self):
        """test auto resource slot multi process"""
        ret_list = AutoBoundedMultiProcess(self.bounded_process, nprocs=2, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+", port=[8000, 8001]
        )
        self.assertEqual(ret_list, self.res)
        ret_list = AutoBoundedMultiProcess(
            self.bounded_process, nprocs=2, use_pool=False, disable_pbar=self.disable_pbar
        ).run(self.a, self.b, self.t, opt="+", port=[8000, 8001])
        self.assertEqual(ret_list, self.res)
        ret_list = AutoBoundedMultiProcess(self.bounded_process, nprocs=1, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+", port=8000
        )
        self.assertEqual(ret_list, self.res)

    def test_multi_thread(self):
        """test multi thread"""
        ret_list = MultiThread(self.process, nworkers=2, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+"
        )
        self.assertEqual(ret_list, self.res)

    def test_auto_multi_thread(self):
        """test auto multi thread"""
        ret_list = AutoMultiThread(self.process, nworkers=2, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+"
        )
        self.assertEqual(ret_list, self.res)
        ret_list = AutoMultiThread(self.process, nworkers=1, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+"
        )
        self.assertEqual(ret_list, self.res)

    def test_async_process(self):
        """test async process"""
        ret_list = AsyncProcess(self.async_process, disable_pbar=self.disable_pbar).run(self.a, self.b, self.t, opt="+")
        self.assertEqual(ret_list, self.res)

    def test_async_multi_process(self):
        """test async multi process"""
        ret_list = AsyncMultiProcess(self.async_process, nprocs=10, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+"
        )
        self.assertEqual(ret_list, self.res)

    def test_auto_async_multi_process(self):
        """test auto async multi process"""
        ret_list = AutoAsyncMultiProcess(self.async_process, nprocs=10, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+"
        )
        self.assertEqual(ret_list, self.res)
        ret_list = AutoAsyncMultiProcess(self.async_process, nprocs=1, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+"
        )
        self.assertEqual(ret_list, self.res)

    def test_parallel_process(self):
        """test parallel process"""
        ret_list = ParallelProcess(self.process_batch, is_single_task_func=False, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+"
        )
        self.assertEqual(ret_list, self.res)

        ret_list = ParallelProcess(
            self.process_batch_v2, is_single_task_func=False, disable_pbar=self.disable_pbar
        ).run(self.a, self.b, self.t, opt="+")
        self.assertEqual(ret_list, self.res)

        ret_list = ParallelProcess(self.process, is_single_task_func=True, disable_pbar=self.disable_pbar).run(
            self.a, self.b, self.t, opt="+"
        )
        self.assertEqual(ret_list, self.res)

        ret_list = ParallelProcess(
            self.process, nprocs=len(self.a), is_single_task_func=True, disable_pbar=self.disable_pbar
        ).run(self.a, self.b, self.t, opt="+")
        self.assertEqual(ret_list, self.res)

        exception = None
        try:
            ret_list = ParallelProcess(
                self.process_batch, is_single_task_func=False, disable_pbar=self.disable_pbar
            ).run(self.a, self.b, self.t, "+")
        except Exception as e:
            exception = e
            print(e)
        self.assertIsInstance(exception, RuntimeError)

    def test_auto_parallel_process(self):
        """test auto parallel process"""
        ret_list = AutoParallelProcess(
            self.process_batch, is_single_task_func=False, nprocs=2, disable_pbar=self.disable_pbar
        ).run(self.a, self.b, self.t, opt="+")
        self.assertEqual(ret_list, self.res)
        ret_list = AutoParallelProcess(
            self.process_batch, is_single_task_func=False, nprocs=1, disable_pbar=self.disable_pbar
        ).run(self.a, self.b, self.t, opt="+")
        self.assertEqual(ret_list, self.res)

        ret_list = AutoParallelProcess(
            self.process_batch_v2, is_single_task_func=False, nprocs=2, disable_pbar=self.disable_pbar
        ).run(self.a, self.b, self.t, opt="+")
        self.assertEqual(ret_list, self.res)
        ret_list = AutoParallelProcess(
            self.process_batch_v2, is_single_task_func=False, nprocs=1, disable_pbar=self.disable_pbar
        ).run(self.a, self.b, self.t, opt="+")
        self.assertEqual(ret_list, self.res)

        ret_list = AutoParallelProcess(
            self.process, is_single_task_func=True, nprocs=2, disable_pbar=self.disable_pbar
        ).run(self.a, self.b, self.t, opt="+")
        self.assertEqual(ret_list, self.res)
        ret_list = AutoParallelProcess(
            self.process, is_single_task_func=True, nprocs=1, disable_pbar=self.disable_pbar
        ).run(self.a, self.b, self.t, opt="+")
        self.assertEqual(ret_list, self.res)


if __name__ == "__main__":
    unittest.main()



import os
import glob
import fitz
import argparse
from lh_tool.iterator import SingleProcess, MultiProcess


def pdf2image(pdf_file, image_file, zoom=1):
    pdf = fitz.open(pdf_file)
    trans = fitz.Matrix(zoom, zoom)

    for index in range(pdf.page_count):
        page = pdf[index]
        pm = page.get_pixmap(matrix=trans)
        if pdf.page_count > 1:
            save_file = image_file.replace(".png", f"-{index + 1:06d}.png")
        else:
            save_file = image_file
        pm.save(save_file)
    pdf.close()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", type=str, help="input pdf file or path", required=True)
    parser.add_argument("-o", "--output", type=str, help="output image file or path")
    parser.add_argument("-z", "--zoom", type=float, default=1, help="zoom image")
    parser.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="convert pdf to image files recursively",
    )
    parser.add_argument("-n", "--nprocs", type=int, default=1, help="number of process")
    opts = parser.parse_args()
    print(opts)

    try:
        input_path = opts.input
        output_path = opts.output
        zoom = opts.zoom
        recursive = opts.recursive
        nprocs = opts.nprocs
        if os.path.isfile(input_path):
            nprocs = 1
            input_file_list = [input_path]
            if output_path is not None:
                output_file_list = [output_path]
            else:
                output_file_list = [os.path.splitext(input_path)[0] + ".png"]
        elif os.path.isdir(input_path):
            assert recursive, "Please use -r to convert pdf to image files recursively."
            input_file_list = glob.glob(os.path.join(input_path, "*.pdf"))
            output_file_list = []
            for input_file in input_file_list:
                if output_path is not None:
                    filename = os.path.basename(input_file)
                    filename = os.path.splitext(filename)[0] + ".png"
                    output_file = os.path.join(output_path, filename)
                else:
                    output_file = os.path.splitext(input_file)[0] + ".png"
                output_file_list.append(output_file)
        else:
            raise ValueError("Input path must be a file or directory.")

        if nprocs > 1:
            iterator = MultiProcess(pdf2image, nprocs=nprocs)
        else:
            iterator = SingleProcess(pdf2image)
        iterator.run(input_file_list, output_file_list, zoom=zoom)
    except Exception as e:
        print(e)


if __name__ == "__main__":
    main()



# -*- encoding: utf-8 -*-
"""
@File    :   monitor_gpu.py
@Time    :   2024/11/08 13:02:38
@Author  :   lihao57
@Version :   1.0
@Contact :   lihao57@baidu.com
"""


import os
import time
import logging
import argparse
import GPUtil
import numpy as np
import pynvml
import threading
from typing import Optional, Union, List, Callable
import functools


class GPUPeakMemoryMonitor:
    """GPU Peak Memory Monitor

    Parameters:
        context (str): The context used to print the time consumption, default is ''
        gpu_ids (Optional[Union[List[int], int]]): The ids of the GPU device that you want to monitor, default is None, meaning all GPUs
        interval (float): The interval between sampling in seconds, default is 1.0
        format_func (Optional[Callable]): The function used to format the output message, default is None
        print_func (Optional[Callable]): The function used to print the output message, default is `print`

    Example:
        ```python
        monitor = GPUPeakMemoryMonitor("block", gpu_ids=0)
        # disable print
        # monitor = GPUPeakMemoryMonitor("block", gpu_ids=0, print_func=None)
        monitor.start()
        # to do something
        peak_memories = monitor.stop()

        # used as context manager
        with GPUPeakMemoryMonitor("block", gpu_ids=None, format_func=lambda x: f"{x / (1024 ** 3):.1f}GB"):
            # to do something

        # used as decorator
        @GPUPeakMemoryMonitor()
        def test():
            # to do something

        test()
        ```
    """

    def __init__(
        self,
        context: str = "",
        gpu_ids: Optional[Union[List[int], int]] = None,
        interval: float = 1.0,
        format_func: Optional[Callable] = None,
        print_func: Optional[Callable] = print,
    ):
        if gpu_ids is None:
            pynvml.nvmlInit()
            gpu_ids = list(range(pynvml.nvmlDeviceGetCount()))
            pynvml.nvmlShutdown()
        elif isinstance(gpu_ids, int):
            gpu_ids = [gpu_ids]
        assert isinstance(gpu_ids, list), f"`gpu_ids` must be a list, but got {type(gpu_ids)}"
        assert len(gpu_ids), "`gpu_ids` cannot be empty"

        self._context = context
        self._gpu_ids = gpu_ids
        self._interval = interval
        self._format_func = format_func
        self._print_func = print_func

        self._peak_memories = [0] * len(self._gpu_ids)
        self._stop_event = threading.Event()
        self._thread: Optional[threading.Thread] = None

    def _monitor(self):
        """monitor"""
        pynvml.nvmlInit()
        handles = [pynvml.nvmlDeviceGetHandleByIndex(gpu_id) for gpu_id in self._gpu_ids]

        try:
            while not self._stop_event.is_set():
                for i, handle in enumerate(handles):
                    mem_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
                    current_memory = mem_info.used
                    if current_memory > self._peak_memories[i]:
                        self._peak_memories[i] = current_memory
                time.sleep(self._interval)
        finally:
            pynvml.nvmlShutdown()

    def start(self):
        """start"""
        if self._thread is not None:
            raise RuntimeError("Monitor is already running")
        self.reset()
        self._thread = threading.Thread(target=self._monitor)
        self._thread.start()

    def stop(self) -> List[int]:
        """stop"""
        if self._thread is None:
            raise RuntimeError("Monitor is not running")
        self._stop_event.set()
        self._thread.join()
        self._thread = None
        self._stop_event.clear()
        peak_memories = self.peak_memories
        if self._print_func is not None:
            if self._context:
                self._print_func(f"{self._context} peak memory usage: {peak_memories}")
            else:
                self._print_func(f"peak memory usage: {peak_memories}")
        return peak_memories

    def reset(self):
        """reset"""
        self._peak_memories = [0] * len(self._gpu_ids)

    @property
    def peak_memories(self) -> List[int]:
        """peak memories"""
        if self._format_func is not None:
            return [self._format_func(x) for x in self._peak_memories]
        else:
            return self._peak_memories

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            """wrapper"""
            if not self._context:
                self._context = str(func).split(" ")[1]
            self.start()
            ret = func(*args, **kwargs)
            self.stop()
            return ret

        wrapper.__name__ = func.__name__
        return wrapper


def monitor(args):
    """
    Monitor gpu utilization

    Args:
        args (dict): arguments

    Returns:
        None
    """
    interval = args.interval
    assert interval > 0, "interval must be greater than 0"

    # get initial gpu utilization and memory usage
    last_utilizations = []
    gpus = GPUtil.getGPUs()
    for gpu in gpus:
        if args.gpus is not None:
            if gpu.id not in args.gpus:
                continue

        last_utilizations.append(gpu.load * 100)
    last_utilizations = np.asarray(last_utilizations)
    assert len(last_utilizations), f"no gpu found {args.gpus}"

    low_utilization_start_time = None
    constant_utilization_start_time = None
    while True:
        trigger = False
        gpus = GPUtil.getGPUs()
        utilizations = []
        for gpu in gpus:
            if args.gpus is not None:
                if gpu.id not in args.gpus:
                    continue

            utilizations.append(gpu.load * 100)
        utilizations = np.asarray(utilizations)
        cur_time = time.time()
        if (utilizations <= args.utilization_threshold).all():
            if low_utilization_start_time is None:
                low_utilization_start_time = cur_time
            else:
                logging.info(f"low utilization detected: {cur_time - low_utilization_start_time}s")
                if cur_time - low_utilization_start_time >= args.time_threshold:
                    trigger = True
                    logging.warning(f"low utilization detected, utilization: {utilizations}")
        else:
            low_utilization_start_time = None

        if args.enable_constant_utilization_monitor:
            if (utilizations == last_utilizations).all():
                if constant_utilization_start_time is None:
                    constant_utilization_start_time = cur_time
                else:
                    logging.info(f"constant utilization detected: {cur_time - constant_utilization_start_time}s")
                    if cur_time - constant_utilization_start_time >= args.time_threshold:
                        trigger = True
                        logging.warning(f"constant utilization detected, utilization: {utilizations}")
            else:
                constant_utilization_start_time = None
            last_utilizations = utilizations

        if trigger:
            if args.command is not None:
                logging.info(f"execute command: {args.command}")
                os.system(args.command)
                logging.info(f"command executed")

            if args.once:
                break

            low_utilization_start_time = None
            constant_utilization_start_time = None

        # sleep
        time.sleep(interval)


def main():
    # set base logging config
    fmt = "[%(asctime)s - %(levelname)s - %(filename)s:%(lineno)s] %(message)s"
    logging.basicConfig(format=fmt, level=logging.INFO)

    parser = argparse.ArgumentParser("GPU Monitor")
    parser.add_argument(
        "-g",
        "--gpus",
        type=int,
        nargs="*",
        help="gpus to be monitored, e.g., --gpus 0 1 2 3 4 5 6 8, default: None (i.e., all gpus)",
    )
    parser.add_argument(
        "-i",
        "--interval",
        type=float,
        help="monitor interval (units of second), default: 10s",
        default=10.0,
    )
    parser.add_argument(
        "-u",
        "--utilization_threshold",
        type=float,
        help="utilization threshold (units of percentage), default: 0%%, negative means no limit",
        default=0,
    )
    parser.add_argument(
        "-e",
        "--enable_constant_utilization_monitor",
        action="store_true",
        help="whether enable constant utilization monitor, default: False",
    )
    parser.add_argument(
        "-t",
        "--time_threshold",
        type=float,
        help="time threshold for monitoring (units of second), default: 300s",
        default=300,
    )
    parser.add_argument(
        "-c",
        "--command",
        type=str,
        help="command to be executed after detecting abnormality, default: None",
    )
    parser.add_argument(
        "-o",
        "--once",
        action="store_true",
        help="only monitor once",
    )

    args = parser.parse_args()
    logging.info(args)

    monitor(args)


if __name__ == "__main__":
    main()



import os
import time
import psutil
import argparse


def startup(pid, period, cmd):
    while psutil.pid_exists(pid):
        time.sleep(period)

    print(f"Start up: {cmd}")
    os.system(cmd)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--pid", type=int, help="pid of waited process")
    parser.add_argument("-c", "--cmd", type=str, help="cmd to be executed")
    parser.add_argument("-p", "--period", type=int, default=600, help="monitoring period")
    opts = parser.parse_args()
    print(opts)

    startup(pid=opts.pid, period=opts.period, cmd=opts.cmd)


if __name__ == "__main__":
    main()



import os
import cv2
import glob
import argparse
from lh_tool.iterator import SingleProcess, MultiProcess
import imageio.v2 as iio


def image2image(input_file, output_file, image_size=None):
    image = iio.imread(input_file)
    if image_size is not None:
        image = cv2.resize(image, tuple(image_size))
    iio.imwrite(output_file, image)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        default=".",
        help="path of input image files",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="path of output image files. The default is the same as the " "path of input image files",
    )
    parser.add_argument(
        "-p",
        "--input_postfix",
        type=str,
        default="png",
        help="original postfix of image filename",
    )
    parser.add_argument(
        "-d",
        "--output_postfix",
        type=str,
        default="jpg",
        help="desired postfix of image filename",
    )
    parser.add_argument("-s", "--size", type=int, nargs=2, help="desired image size")
    parser.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="convert video to images recursively",
    )
    parser.add_argument("-n", "--nprocs", type=int, default=1, help="number of process")
    opts = parser.parse_args()
    print(opts)

    try:
        input_path = opts.input
        output_path = opts.output
        input_postfix = opts.input_postfix
        output_postfix = opts.output_postfix
        image_size = None if opts.size is None else tuple(opts.size)
        recursive = opts.recursive
        nprocs = opts.nprocs
        if recursive:
            input_file_list = glob.glob(
                os.path.join(input_path, "**", f"*.{input_postfix}"),
                recursive=True,
            )
        else:
            input_file_list = glob.glob(os.path.join(input_path, f"*.{input_postfix}"))
        output_file_list = []
        for input_file in input_file_list:
            if output_path is not None:
                filename = os.path.basename(input_file)
                filename = os.path.splitext(filename)[0] + f".{output_postfix}"
                output_file = os.path.join(output_path, filename)
            else:
                output_file = os.path.splitext(input_file)[0] + f".{output_postfix}"
            output_file_list.append(output_file)

        if nprocs == 1:
            iterator = SingleProcess(image2image)
        else:
            iterator = MultiProcess(image2image, nprocs=nprocs)
        iterator.run(input_file_list, output_file_list, image_size=image_size)

    except AssertionError as e:
        print(e)


if __name__ == "__main__":
    main()



import os
import xlrd
import argparse


def get_cell_font_bold(workbook, cell):
    xf_index = cell.xf_index
    xf_style = workbook.xf_list[xf_index]
    font_index = xf_style.font_index
    font = workbook.font_list[font_index]
    return font.bold


def excel2markdown(excel_file, markdown_file, sheet_index=0, style=""):
    markdown_file = os.path.splitext(excel_file)[0] + ".md" if markdown_file is None else markdown_file

    workbook = xlrd.open_workbook(excel_file, formatting_info=True)
    sheet = workbook.sheet_by_index(sheet_index)

    with open(markdown_file, "w") as f:
        f.write("<html>\n")
        f.write('<table align="center">\n')

        merged_cells = dict()
        for (rlow, rhigh, clow, chigh) in sheet.merged_cells:
            merged_cells[(rlow, clow)] = (rhigh - rlow, chigh - clow)

        precisions = []
        for j in range(sheet.ncols):
            max_precision = 0
            for i in range(sheet.nrows):
                if sheet.cell_type(i, j) == 2:
                    value = sheet.cell_value(i, j)
                    if value == round(value):
                        precision = 0
                    else:
                        precision = len(str(value).split(".")[-1])
                    max_precision = max(max_precision, precision)
            precisions.append(max_precision)

        for i in range(sheet.nrows):
            f.write(f"\t<tr{style}>\n")

            for j in range(sheet.ncols):
                cell = sheet.cell(i, j)
                type = cell.ctype
                value = cell.value
                bold = get_cell_font_bold(workbook, cell)
                rowspan = ""
                colspan = ""
                if (i, j) in merged_cells.keys():
                    (rows, cols) = merged_cells[(i, j)]
                    rowspan = f' rowspan="{rows}"'
                    colspan = f' colspan="{cols}"'
                if value != "":
                    if type == 2:
                        precision = precisions[j]
                        value = f"%.{precision}f" % value
                    if bold:
                        f.write(f'\t\t<td{rowspan}{colspan} align="center"><b>{value}</b></td>\n')
                    else:
                        f.write(f'\t\t<td{rowspan}{colspan} align="center">{value}</td>\n')
            f.write("\t</tr>\n")
        f.write("</table>\n")
        f.write("</html>\n")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", type=str, help="input excel file", required=True)
    parser.add_argument("-o", "--output", type=str, help="output markdown file")
    parser.add_argument("-s", "--sheet_index", type=int, default=0, help="sheet index")
    opts = parser.parse_args()
    print(opts)

    excel_file = opts.input
    markdown_file = opts.output
    sheet_index = opts.sheet_index
    style = ' style = "font-size:10px"'
    excel2markdown(excel_file, markdown_file, sheet_index, style)


if __name__ == "__main__":
    main()



# -*- encoding: utf-8 -*-
"""
@File    :   timer.py
@Time    :   2024/03/09 17:04:14
@Author  :   lihao57
@Version :   1.0
@Contact :   lihao57@baidu.com
"""


import threading
import time
from functools import partial
from typing import Callable, Iterable, Mapping


class Timer:
    """
    Timer

    Parameters:
        interval (int): interval time (unit: ms), default 1000ms
        callback (callable): callback function
        args (Optinal): callback args
        kwargs (Optinal): callback kwargs

    Example:
        ```python
        start_time = time.time()

        def callback():
            print(int((time.time() - start_time) * 1000))
            time.sleep(0.05)

        timer = Timer(100, callback)
        timer.start()
        time.sleep(1)
        timer.stop()

        # used as context manager
        with Timer(100, callback):
            time.sleep(1)
        ```
    """

    def __init__(
        self,
        interval: int = 1000,
        callback: Callable = None,
        args: Iterable = (),
        kwargs: Mapping = {},
    ):
        assert interval > 0, "interval must be greater than 0"
        assert isinstance(callback, Callable), "callback must be callable"
        self.interval = interval
        self.args = args
        self.kwargs = kwargs
        self.callback = partial(callback, *args, **kwargs)

        self.lock = threading.Lock()
        self._set_running(False)

    def start(self, session=None):
        """start timer"""
        assert not self.is_running(), "timer is start, can not start"
        self._set_running(True)
        self.thread = threading.Thread(target=self._run)
        if session is not None:
            session.register_thread(self.thread)
        self.thread.start()

    def stop(self):
        """stop timer"""
        assert self.is_running(), "timer is not start, can not stop"
        self._set_running(False)
        self.thread.join()

    def is_running(self):
        """check timer is running or not"""
        self.lock.acquire()
        is_running = self._is_running
        self.lock.release()
        return is_running

    def _set_running(self, is_running: bool):
        """set running status"""
        self.lock.acquire()
        self._is_running = is_running
        self.lock.release()

    def set_interval(self, interval: int):
        """set interval"""
        assert interval > 0, "interval must be greater than 0"
        assert not self.is_running(), "timer is start, can not set interval"
        self.interval = interval

    def _run(self):
        """run timer"""
        start_time = int(time.time() * 1000)
        cnt = 0
        while True:
            if not self.is_running():  # timer is stop
                break

            cnt += 1
            now_time = int(time.time() * 1000)
            sleep_time = cnt * self.interval - (now_time - start_time)
            if sleep_time > 0:
                time.sleep(sleep_time / 1000)

            # run callback
            self.callback()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()



import os
import cv2
import glob
import tqdm
import argparse
from lh_tool.iterator import SingleProcess, MultiProcess


def video2images(video_file, image_path, postfix, frameSize=None):
    assert os.path.isfile(video_file), f"'{video_file}' is not a file"
    image_path = os.path.splitext(video_file)[0] if image_path is None else image_path
    os.makedirs(image_path, exist_ok=True)

    videoCapture = cv2.VideoCapture(video_file)
    assert videoCapture.isOpened(), f"Failed to open file: '{video_file}'"

    frame_count = int(videoCapture.get(cv2.CAP_PROP_FRAME_COUNT))
    for index in tqdm.trange(frame_count, desc=video_file):
        status, image = videoCapture.read()
        if not status:
            break
        if frameSize is not None:
            image = cv2.resize(image, tuple(frameSize))
        image = image[:, :, ::-1]  # BGR2RGB
        image_file = os.path.join(image_path, f"{index + 1:06d}.{postfix}")
        cv2.imwrite(image_file, image)

    videoCapture.release()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        default=".",
        help="input video file or input path",
    )
    parser.add_argument("-o", "--output", type=str, help="output path")
    parser.add_argument(
        "-p",
        "--postfix",
        type=str,
        default="png",
        help="postfix of image filename",
    )
    parser.add_argument("-s", "--size", type=int, nargs=2, help="desired frame size for image")
    parser.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="convert video to images recursively",
    )
    parser.add_argument("-n", "--nprocs", type=int, default=1, help="number of process")
    opts = parser.parse_args()
    print(opts)

    try:
        video_file = opts.input
        image_path = opts.output
        postfix = opts.postfix
        size = None if opts.size is None else tuple(opts.size)
        recursive = opts.recursive
        nprocs = opts.nprocs
        if recursive:
            video_file_list = glob.glob(os.path.join(opts.input, "*.mp4"))
            if nprocs == 1:
                iterator = SingleProcess(video2images)
            else:
                iterator = MultiProcess(video2images, nprocs=nprocs)
            iterator.run(
                video_file_list,
                image_path=None,
                postfix=postfix,
                frameSize=size,
            )
        else:
            video2images(video_file, image_path, postfix, size)

    except AssertionError as e:
        print(e)


if __name__ == "__main__":
    main()



# -*- encoding: utf-8 -*-
"""
@File    :   pcd2ply.py
@Time    :   2024/09/16 20:28:50
@Author  :   lh9171338
@Version :   1.0
@Contact :   2909171338@qq.com
"""


import os
import glob
import cv2
import numpy as np
import open3d as o3d
from pypcd import pypcd
import argparse
from lh_tool.iterator import SingleProcess, MultiProcess


def get_class_colors(num_classes):
    """
    Get class colors

    Args:
        num_classes (int): number of classes

    Returns:
        colors (list): list of class colors
    """
    colors = np.ones((1, num_classes, 3), dtype="float32")
    colors[:, :, 0] = np.linspace(0, 360, num_classes + 1)[:-1]
    colors = cv2.cvtColor(colors, cv2.COLOR_HSV2RGB)
    colors = colors[0]

    return colors


def pcd2ply(
    pcd_file,
    ply_file,
    with_label=False,
):
    """
    Convert pcd file to ply file

    Args:
        pcd_file (str): pcd file path
        ply_file (str): ply file path
        with_label (bool, optional): Whether to include label. Defaults to False

    Returns:
        None
    """
    pc = pypcd.PointCloud.from_path(pcd_file)
    pc_data = pc.pc_data
    points = np.stack(
        [
            pc_data["x"],
            pc_data["y"],
            pc_data["z"],
        ],
        axis=-1,
    )
    if with_label and "label" in pc.fields:
        labels = pc_data["label"]
        unique_labels, labels = np.unique(labels, return_inverse=True)
        num_classes = len(unique_labels)
        class_colors = get_class_colors(num_classes)
        colors = class_colors[labels]
    else:
        colors = np.ones_like(points)

    point_cloud = o3d.geometry.PointCloud()
    point_cloud.points = o3d.utility.Vector3dVector(points)
    point_cloud.colors = o3d.utility.Vector3dVector(colors)
    o3d.io.write_point_cloud(ply_file, point_cloud)


def main():
    """main"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        default=".",
        help="path of input pcd files",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="path of output ply files. The default is the same as the " "path of input pcd files",
    )
    parser.add_argument(
        "-l",
        "--label",
        dest="with_label",
        action="store_true",
        help="Whether to include label",
    )
    parser.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="convert pcd to ply files recursively",
    )
    parser.add_argument("-n", "--nprocs", type=int, default=1, help="number of process")
    opts = parser.parse_args()
    print(opts)

    try:
        input_path = opts.input
        output_path = opts.output
        with_label = opts.with_label
        recursive = opts.recursive
        nprocs = opts.nprocs
        if os.path.isfile(input_path):
            nprocs = 1
            input_file_list = [input_path]
            if output_path is not None:
                output_file_list = [output_path]
            else:
                output_file_list = [os.path.splitext(input_path)[0] + ".ply"]
        elif os.path.isdir(input_path):
            assert recursive, "Please use -r to convert pcd to ply files recursively."
            input_file_list = glob.glob(os.path.join(input_path, "*.pcd"))
            output_file_list = []
            for input_file in input_file_list:
                if output_path is not None:
                    filename = os.path.basename(input_file)
                    filename = os.path.splitext(filename)[0] + ".ply"
                    output_file = os.path.join(output_path, filename)
                else:
                    output_file = os.path.splitext(input_file)[0] + ".ply"
                output_file_list.append(output_file)
        else:
            raise ValueError("Input path must be a file or directory.")

        if nprocs > 1:
            iterator = MultiProcess(pcd2ply, nprocs=nprocs)
        else:
            iterator = SingleProcess(pcd2ply)
        iterator.run(input_file_list, output_file_list, with_label=with_label)
    except Exception as e:
        print(e)


if __name__ == "__main__":
    main()



# -*- encoding: utf-8 -*-
"""
@File    :   test_timer.py
@Time    :   2024/03/14 13:29:57
@Author  :   lihao57
@Version :   1.0
@Contact :   lihao57@baidu.com
"""

import unittest
import time
import numpy as np
import sys

sys.path.append("../src")
from lh_tool.timer import Timer


class TestTimer(unittest.TestCase):
    """TestTimer"""

    @staticmethod
    def callback(times):
        """callback"""
        times.append(time.time())

    def test_timer(self):
        """test timer"""
        times = [time.time()]
        timer = Timer(100, self.callback, (times,))
        timer.start()
        time.sleep(1)
        timer.stop()
        times = np.array(times)
        times = times[1:] - times[:-1]
        error = np.abs(times - 0.1)
        print(error)
        # 误差小于1ms
        self.assertTrue(error.max() < 0.001)

        times = [time.time()]
        with Timer(100, self.callback, (times,)):
            time.sleep(1)
        times = np.array(times)
        times = times[1:] - times[:-1]
        error = np.abs(times - 0.1)
        print(error)
        # 误差小于1ms
        self.assertTrue(error.max() < 0.001)


if __name__ == "__main__":
    unittest.main()
